\name{lavaan-deprecated}
\alias{lavaan-deprecated}
\title{Deprecated functions in lavaan package}
\details{
The functions listed in the first column below are deprecated; calls to them
should be replaced by calls to the corresponding function in the second column.
\tabular{ll}{
#  bootstrapLavaan \tab lavBootstrap \cr
  char2num \tab lav_char2num \cr
  cor2cov \tab lav_cor2cov \cr
  getCov \tab lav_getcov \cr
  inspectSampleCov \tab lavInspectSampleCov \cr
#  lavaanNames \tab lavNames \cr
  mplus2lavaan.modelSyntax \tab lav_mplus_syntax_model \cr
  mplus2lavaan \tab lav_Mplus_lavaan \cr
  parameterestimates \tab lavParameterEstimates \cr
#  simulateData \tab lavSimulateData \cr
}
}
\description{
Deprecated functions in lavaan.
}
